from sportstat.FBRef import FBRef
from sportstat.Understat import Understat
from sportstat.FiveThirtyEight import FiveThirtyEight
from IPython.display import clear_output
import pandas as pd
from sportstat.shared_functions import check_season


class ScraperFC:
    
    def __init__(self):
        self.fbref = FBRef()
        self.understat = Understat()
        self.fte = FiveThirtyEight()
        
        
    def close(self):
        self.fbref.close()
        self.understat.close()
        self.fte.close()
        
    
    def scrape_matches(self, year, league, save=False):
        if not check_season(year,league,'All'):
            return -1
        
#         print('Preparing to gather match data from Understat and FBRef.')
        season = str(year-1)+'-'+str(year)
#         num_matches = 380
#         # get FBREF links
#         fbref_links = self.fbref.get_match_links(year,league)
#         assert len(fbref_links) == num_matches
#         # get Understat links
#         understat_links = self.understat.get_match_links(year, league)
#         assert len(understat_links) == num_matches
#         clear_output()
        
#         fbref_cols = ['Date','Home Team','Away Team','Home Goals','Away Goals',
#                       'Home Ast','Away Ast','FBRef Home xG','FBRef Away xG','Home npxG',
#                       'Away npxG','FBRef Home xA','FBRef Away xA','Home psxG','Away psxG']
#         fbref_df = pd.DataFrame(columns=fbref_cols)
#         understat_cols = ['Date','Home Team','Away Team','Home Goals','Away Goals',
#                           'Home Ast','Away Ast','Understat Home xG','Understat Away xG',
#                           'Understat Home xA','Understat Away xA','Home xPts','Away xPts']
#         understat_df = pd.DataFrame(columns=understat_cols)

        print('Scraping match data from FiveThirtyEight for the '+season+" "+league+' season.')
        fte_df = self.fte.scrape_matches(year, league)
        clear_output()
        
        print('Scraping match data from FBRef for the '+season+" "+league+' season.')
        fbref_df = self.fbref.scrape_matches(year, league)
        clear_output()
        
        print('Scraping match data from Understat for the '+season+" "+league+' season.')
        understat_df = self.understat.scrape_matches(year,league)
        clear_output()
        
#         for i in range(num_matches):
#             print('Scraping match ' + str(i+1) + '/' + str(num_matches) + ' in the ' + season + ' EPL season.')
            
#             fbref_link = fbref_links[i]
#             understat_link = understat_links[i]
            
#             fbref_match = self.fbref.scrape_match(fbref_link,'EPL')
#             understat_match = self.understat.scrape_match(understat_link)
            
#             fbref_df = fbref_df.append(fbref_match, ignore_index=True)
#             understat_df = understat_df.append(understat_match, ignore_index=True)
            
#             clear_output()

        # merge dataframes
        df = fbref_df.merge(understat_df, how='outer',
                            left_index=True, right_index=True,
                            on=['Date','Home Team','Away Team',
                                'Home Goals','Away Goals',
                                'Home Ast','Away Ast'])
        df = df.merge(fte_df, how='outer',
                      left_index=True, right_index=True,
                      on=['Date','Home Team','Away Team'])
        
        # save to CSV if requested by user
        if save:
            filename = season+"_"+league+"_matches.csv"
            df.to_csv(path_or_buf=filename, index=False)
            print('Matches dataframe saved to ' + filename)
            return filename
        else:
            return df
        
            


import datetime
from IPython.display import clear_output
import numpy as np
import pandas as pd
from sportstat.shared_functions import check_season
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from webdriver_manager.chrome import ChromeDriverManager
from urllib.request import urlopen
import requests
from bs4 import BeautifulSoup
from lxml import etree
from tqdm import tqdm
import time


class FBRef:
    
    ################################################################################
    def __init__(self):
        options = Options()
        options.headless = True
        options.add_argument("window-size=1400,600")
        options.add_argument("user-agent=Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.0.0 Safari/537.36")
        self.driver = webdriver.Chrome(ChromeDriverManager().install(), options=options)
      
    ################################################################################    
    def close(self):
        self.driver.close()
        self.driver.quit()

    ################################################################################
    def get_season_link(self, year, league):
        if league == 'EPL':
            url = 'https://fbref.com/en/comps/9/history/Premier-League-Seasons'
            if year >= 2008:
                finder = 'Premier-League-Stats'
            else:
                finder = 'Premiership-Stats'
        elif league == 'La Liga':
            url = 'https://fbref.com/en/comps/12/history/La-Liga-Seasons'
            finder = 'La-Liga-Stats'
        elif league == "Bundesliga":
            url = "https://fbref.com/en/comps/20/history/Bundesliga-Seasons"
            finder = "Bundesliga-Stats"
        elif league == "Serie A":
            url = "https://fbref.com/en/comps/11/history/Serie-A-Seasons"
            finder = "Serie-A-Stats"
        elif league == "Ligue 1":
            url = "https://fbref.com/en/comps/13/history/Ligue-1-Seasons"
            if year >= 2003:
                finder = 'Ligue-1'
            else:
                finder = 'Division-1'
        elif league == "MLS":
            url = "https://fbref.com/en/comps/22/history/Major-League-Soccer-Seasons"
            finder = "Major-League-Soccer-Stats"
        else:
            print('ERROR: League not found. Options are \"EPL\", \"La Liga\", '+
                  '\"Bundesliga\", \"Serie A\", \"Ligue 1\"')
            return -1
        self.driver.get(url)
        if league != "MLS":
            season = str(year-1)+'-'+str(year)
        else:
            season = str(year)
        for element in self.driver.find_elements_by_link_text(season):
            if finder in element.get_attribute('href'):
                return element.get_attribute('href')
            else:
                print('ERROR: Season not found.')
                return -1
    
    ################################################################################
    def get_match_links(self, year, league):
        print("Gathering match links.")
        url = self.get_season_link(year, league)
        
        # go to the scores and fixtures page
        url = url.split('/')
        first_half = '/'.join(url[:-1])
        second_half = url[-1].split('-')
        second_half = '-'.join(second_half[:-1])+'-Score-and-Fixtures'
        url = first_half+'/schedule/'+second_half
        self.driver.get(url)
        
        # Get links to all of the matches in that season
        if league == 'EPL':
            if year >= 2008:
                finder = '-Premier-League'
            else:
                finder = '-Premiership'
        elif league == 'La Liga':
            finder = '-La-Liga'
        elif league == 'Bundesliga':
            finder = '-Bundesliga'
        elif league == 'Serie A':
            finder = '-Serie-A'
        elif league == 'Ligue 1':
            if year >= 2003:
                finder = '-Ligue-1'
            else:
                finder = '-Division-1'
        elif league == "MLS":
            finder = "-Major-League-Soccer"
        else:
            print('ERROR: League not found. Options are \"EPL\", \"La Liga\", '+
                  '\"Bundesliga\", \"Serie A\", \"Ligue 1\", \"MLS\".')
            return -1
        
        links = set()
        # only get match links from the fixtures table
        for table in self.driver.find_elements(By.TAG_NAME, "table"):
            if table.get_attribute('id')!='' and table.get_attribute('class')!='':
                # find the match links
                for element in tqdm(table.find_elements(By.TAG_NAME, "a")):
                    href = element.get_attribute('href')
                    if (href) and ("/matches/" in href) and (href.endswith(finder)):
                        links.add(href)
        
        return list(links)
    
    ################################################################################
    def add_team_ids(self, df, insert_index, url, tag_name):
        self.driver.get(url)
        team_ids = list()
#         for el in self.driver.find_elements_by_xpath('//{}[@data-stat="squad"]'.format(tag_name)):
        for el in self.driver.find_elements(By.XPATH, '//{}[@data-stat="squad"]'.format(tag_name)):
            if el.text != '' and el.text != 'Squad':
                team_id = el.find_element(By.TAG_NAME, 'a') \
                    .get_attribute('href') \
                    .split('/squads/')[-1] \
                    .split('/')[0]
                team_ids.append(team_id)
        df.insert(insert_index, 'team_id', team_ids)
        return df

    ################################################################################
    def add_player_ids_and_links(self, df, url):
        self.driver.get(url)
        player_ids = list()
        player_links = list()
        for el in self.driver.find_elements_by_xpath('//td[@data-stat="player"]'):
            if el.text != '' and el.text != 'Player':
                player_id = el.find_element(By.TAG_NAME, 'a') \
                    .get_attribute('href') \
                    .split('/players/')[-1] \
                    .split('/')[0]
                player_ids.append(player_id)
                player_links.append(el.find_element(By.TAG_NAME, 'a').get_attribute('href'))
        df.insert(2, 'player_id', player_ids) # insert player IDs as new col in df
        df.insert(2, 'player_link', player_links) # insert player links as new col in df
        return df
    
    ################################################################################
    def normalize_table(self, xpath):
        button = self.driver.find_element(By.XPATH, xpath)
        self.driver.execute_script("arguments[0].click()",button)
    
    ################################################################################
    def get_html_w_id(self, ID):
        return self.driver.find_element(By.ID, ID).get_attribute('outerHTML')

    ################################################################################
    def scrape_league_table(self, year, league, normalize=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        print('Scraping {} {} league table'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        df = pd.read_html(url)
        lg_tbl = df[0].copy()
        #### Drop columns and normalize
        if year >= 2018:
            lg_tbl.drop(columns=["xGD/90"], inplace=True)
        if normalize and year >= 2018:
            lg_tbl.iloc[:,3:13] = lg_tbl.iloc[:,3:13].divide(lg_tbl["MP"], axis="rows")
        elif normalize and year < 2018:
            lg_tbl.iloc[:,3:10] = lg_tbl.iloc[:,3:10].divide(lg_tbl["MP"], axis="rows")
        #### Scrape western conference if MLS
        if league == "MLS":
            west_tbl = df[2].copy()
            if year >= 2018:
                west_tbl.drop(columns=["xGD/90"], inplace=True)
            if normalize and year >= 2018:
                west_tbl.iloc[:,3:13] = west_tbl.iloc[:,3:13].divide(west_tbl["MP"], axis="rows")
            elif normalize and year < 2018:
                west_tbl.iloc[:,3:10] = west_tbl.iloc[:,3:10].divide(west_tbl["MP"], axis="rows")
            return (lg_tbl, west_tbl)
        lg_tbl = self.add_team_ids(lg_tbl, 2, url, 'td') # Get team IDs
        return lg_tbl
    
    ################################################################################
    def scrape_standard(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        print('Scraping {} {} standard stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/stats/' + new[-1]
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table('//*[@id=\"stats_standard_per_match_toggle\"]')
            # get html and scrape table
            html = self.get_html_w_id("stats_standard")
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            df.drop(columns="Per 90 Minutes", level=0, inplace=True)
            df.drop(columns="Matches", level=1, inplace=True)
            # convert some column types from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            # add some calculated columns
            df[("Performance","G+A")] = df[("Performance","Gls")] + df[("Performance","Ast")]
            df[("Performance","G+A-PK")] = df[("Performance","G+A")] - df[("Performance","PK")]
            if year >= 2018:
                df[("Expected","xG+xA")] = df[("Expected","xG")] + df[("Expected","xA")]
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            drop_cols = squad.xs("Per 90 Minutes", axis=1, level=0, drop_level=False).columns
            squad.drop(columns=drop_cols, inplace=True)
            vs.drop(columns=drop_cols, inplace=True)
            if normalize:
                squad.iloc[:,8:] = squad.iloc[:,8:].divide(squad[("Playing Time","90s")], axis="rows")
                vs.iloc[:,8:] = vs.iloc[:,8:].divide(vs[("Playing Time","90s")], axis="rows")
            col = ("Performance","G+A")
            squad[col] = squad[("Performance","Gls")] + squad[("Performance","Ast")]
            vs[col] = vs[("Performance","Gls")] + vs[("Performance","Ast")]
            col = ("Performance","G+A-PK")
            squad[col] = squad[("Performance","G+A")] - squad[("Performance","PK")]
            vs[col] = vs[("Performance","G+A")] - vs[("Performance","PK")]
            if year >= 2018:
                col = ("Expected","xG+xA")
                squad[col] = squad[("Expected","xG")] + squad[("Expected","xA")]
                vs[col] = vs[("Expected","xG")] + vs[("Expected","xA")]
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
    
    ################################################################################
    def scrape_gk(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        elif league in['La Liga','Bundesliga','Ligue 1'] and year<2000:
            print('Goalkeeping stats not available from',league,'before 1999/2000 season.')
            return -1
        elif league=='Serie A' and year<1999:
            print('Goalkeeping stats not available from Serie A before 1998/99 season.')
            return -1
        print('Scraping {} {} goalkeeping stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/keepers/' + new[-1]
        new = new.replace("https","http")
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table("//*[@id=\"stats_keeper_per_match_toggle\"]")
            # get html and scrape table
            html = self.get_html_w_id("stats_keeper")
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            df.drop(columns=("Performance","GA90"), inplace=True)
            df.drop(columns="Matches", level=1, inplace=True)
            # convert type from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            squad.drop(columns=("Performance","GA90"), inplace=True)
            vs.drop(columns=("Performance","GA90"), inplace=True)
            if normalize:
                keep_cols = [("Performance","Save%"), ("Performance","CS%"), ("Penalty Kicks","Save%")]
                keep = squad[keep_cols]
                squad.iloc[:,6:] = squad.iloc[:,6:].divide(squad[("Playing Time","90s")], axis="rows")
                squad[keep_cols] = keep
                keep = vs[keep_cols]
                vs.iloc[:,6:] = vs.iloc[:,6:].divide(vs[("Playing Time","90s")], axis="rows")
                vs[keep_cols] = keep
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
    
    ################################################################################
    def scrape_adv_gk(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        elif year < 2018:
            print("Advanced goalkeeping stats not available from before the 2017/18 season.")
            return -1
        print('Scraping {} {} advanced goalkeeping stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/keepersadv/' + new[-1]
        new = new.replace("https","http")
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table("//*[@id=\"stats_keeper_adv_per_match_toggle\"]")
            # get html and scrape table
            html = self.get_html_w_id("stats_keeper_adv")
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            df.drop(columns=["Matches", "#OPA/90", "/90"], level=1, inplace=True)
            # convert type from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            squad.drop(columns=[("Expected","/90"), ("Sweeper","#OPA/90")], inplace=True)
            vs.drop(columns=[("Expected","/90"), ("Sweeper","#OPA/90")], inplace=True)
            if normalize:
                keep_cols = [("Launched","Cmp%"),
                             ("Passes","Launch%"), ("Passes","AvgLen"),
                             ("Goal Kicks","Launch%"), ("Goal Kicks", "AvgLen"), 
                             ("Crosses","Stp%"), ("Sweeper","AvgDist")]
                keep = squad[keep_cols]
                squad.iloc[:,3:] = squad.iloc[:,3:].divide(squad[("Unnamed: 2_level_0","90s")], axis="rows")
                squad[keep_cols] = keep
                keep = vs[keep_cols]
                vs.iloc[:,3:] = vs.iloc[:,3:].divide(vs[("Unnamed: 2_level_0","90s")], axis="rows")
                vs[keep_cols] = keep
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
    
    ################################################################################
    def scrape_shooting(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        print('Scraping {} {} shooting stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/shooting/' + new[-1]
        new = new.replace("https","http")
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table("//*[@id=\"stats_shooting_per_match_toggle\"]")
            # get html and scrape table
            html = self.get_html_w_id("stats_shooting")
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            df.drop(columns=[("Standard","Sh/90"),("Standard","SoT/90")], inplace=True)
            df.drop(columns="Matches", level=1, inplace=True)
            # convert type from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            squad.drop(columns=[("Standard","Sh/90"), ("Standard","SoT/90")], 
                       inplace=True)
            vs.drop(columns=[("Standard","Sh/90"), ("Standard","SoT/90")],
                    inplace=True)
            if normalize:
                keep_cols = [("Standard","SoT%"), ("Standard","Dist")]
                keep = squad[keep_cols]
                squad.iloc[:,3:] = squad.iloc[:,3:].divide(squad[("Unnamed: 2_level_0","90s")], axis="rows")
                squad[keep_cols] = keep
                keep = vs[keep_cols]
                vs.iloc[:,3:] = vs.iloc[:,3:].divide(vs[("Unnamed: 2_level_0","90s")], axis="rows")
                vs[keep_cols] = keep
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
    
    ################################################################################
    def scrape_passing(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        elif year < 2018:
            print("Passing stats not available from before the 2017/18 season.")
            return -1
        print('Scraping {} {} passing stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/passing/' + new[-1]
        new = new.replace("https","http")
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table("//*[@id=\"stats_passing_per_match_toggle\"]")
            # get html and scrape table
            html = self.get_html_w_id('stats_passing')
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            df.drop(columns=[("Unnamed: 30_level_0","Matches")],
                    inplace=True)
            # convert type from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            if normalize:
                keep_cols = [("Total","Cmp%"), ("Short","Cmp%"), ("Medium","Cmp%"), ("Long","Cmp%")]
                keep = squad[keep_cols]
                squad.iloc[:,3:] = squad.iloc[:,3:].divide(squad[("Unnamed: 2_level_0","90s")], axis="rows")
                squad[keep_cols] = keep
                keep = vs[keep_cols]
                vs.iloc[:,3:] = vs.iloc[:,3:].divide(vs[("Unnamed: 2_level_0","90s")], axis="rows")
                vs[keep_cols] = keep
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
    
    ################################################################################
    def scrape_passing_types(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        elif year < 2018:
            print("Passing type stats not available from before the 2017/18 season.")
            return -1
        print('Scraping {} {} passing type stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/passing_types/' + new[-1]
        new = new.replace("https","http")
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table("//*[@id=\"stats_passing_types_per_match_toggle\"]")
            # get html and scrape table
            html = self.get_html_w_id('stats_passing_types')
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            df.drop(columns=[("Unnamed: 33_level_0","Matches")],
                    inplace=True)
            # convert type from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            if normalize:
                squad.iloc[:,3:] = squad.iloc[:,3:].divide(squad[("Unnamed: 2_level_0","90s")], axis="rows")
                vs.iloc[:,3:] = vs.iloc[:,3:].divide(vs[("Unnamed: 2_level_0","90s")], axis="rows")
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
       
    ################################################################################
    def scrape_goal_shot_creation(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        elif year < 2018:
            print("Goal and shot creation stats not available from before the 2017/18 season.")
            return -1
        print('Scraping {} {} goal and shot creation stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/gca/' + new[-1]
        new = new.replace("https","http")
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table("//*[@id=\"stats_gca_per_match_toggle\"]")
            # get html and scrape table
            html = self.get_html_w_id('stats_gca')
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            # df.drop(columns=[("SCA","SCA90"), ("GCA","GCA90")], inplace=True)
            df.drop(columns=["SCA90", "GCA90", "Matches"], level=1, inplace=True)
            # convert type from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            squad.drop(columns=[("SCA","SCA90"), ("GCA","GCA90")], inplace=True)
            vs.drop(columns=[("SCA","SCA90"), ("GCA","GCA90")], inplace=True)
            if normalize:
                squad.iloc[:,3:] = squad.iloc[:,3:].divide(squad[("Unnamed: 2_level_0","90s")], axis="rows")
                vs.iloc[:,3:] = vs.iloc[:,3:].divide(vs[("Unnamed: 2_level_0","90s")], axis="rows")
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
    
    ################################################################################
    def scrape_defensive(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        elif year < 2018:
            print("Defensive stats not available from before the 2017/18 season.")
            return -1
        print('Scraping {} {} defending stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/defense/' + new[-1]
        new = new.replace("https","http")
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table("//*[@id=\"stats_defense_per_match_toggle\"]")
            # get html and scrape table
            html = self.get_html_w_id('stats_defense')
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            df.drop(columns=[("Unnamed: 31_level_0","Matches")],
                    inplace=True)
            # convert type from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            if normalize:
                keep_cols = [("Vs Dribbles","Tkl%"), ("Pressures","%")]
                keep = squad[keep_cols]
                squad.iloc[:,3:] = squad.iloc[:,3:].divide(squad[("Unnamed: 2_level_0","90s")], axis="rows")
                squad[keep_cols] = keep
                keep = vs[keep_cols]
                vs.iloc[:,3:] = vs.iloc[:,3:].divide(vs[("Unnamed: 2_level_0","90s")], axis="rows")
                vs[keep_cols] = keep
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
    
    ################################################################################
    def scrape_possession(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        elif year < 2018:
            print("Possession stats not available from before the 2017/18 season.")
            return -1
        print('Scraping {} {} possession stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/possession/' + new[-1]
        new = new.replace("https","http")
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table("//*[@id=\"stats_possession_per_match_toggle\"]")
            # get html and scrape table
            html = self.get_html_w_id('stats_possession')
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            df.drop(columns=[("Unnamed: 32_level_0","Matches")],
                    inplace=True)
            # convert type from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            if normalize:
                keep_cols = [("Dribbles","Succ%"),("Receiving","Rec%")]
                keep = squad[keep_cols]
                squad.iloc[:,4:] = squad.iloc[:,4:].divide(squad[("Unnamed: 3_level_0","90s")], axis="rows")
                squad[keep_cols] = keep
                keep = vs[keep_cols]
                vs.iloc[:,4:] = vs.iloc[:,4:].divide(vs[("Unnamed: 3_level_0","90s")], axis="rows")
                vs[keep_cols] = keep
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
    
    ################################################################################
    def scrape_playing_time(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        print('Scraping {} {} playing time stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/playingtime/' + new[-1]
        new = new.replace("https","http")
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table("//*[@id=\"stats_playing_time_per_match_toggle\"]")
            # get html and scrape table
            html = self.get_html_w_id('stats_playing_time')
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            df.drop(columns=("Team Success","+/-90"), inplace=True)
            df.drop(columns="Matches", level=1, inplace=True)
            if year >= 2018:
                df.drop(columns="xG+/-90", level=1, inplace=True)
            # convert type from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            squad.drop(columns=("Team Success","+/-90"), inplace=True)
            vs.drop(columns=("Team Success","+/-90"), inplace=True)
            if year >= 2018:
                squad.drop(columns=("Team Success (xG)","xG+/-90"), inplace=True)
                vs.drop(columns=("Team Success (xG)","xG+/-90"), inplace=True)
            if normalize:
                keep_cols = [("Playing Time","Mn/MP"), ("Playing Time","Min%"),
                             ("Playing Time","90s"), ("Starts","Mn/Start")]
                keep = squad[keep_cols]
                squad.iloc[:,4:] = squad.iloc[:,4:].divide(squad[("Playing Time","MP")], axis="rows")
                squad[keep_cols] = keep
                keep = vs[keep_cols]
                vs.iloc[:,4:] = vs.iloc[:,4:].divide(vs[("Playing Time","MP")], axis="rows")
                vs[keep_cols] = keep
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
    
    ################################################################################
    def scrape_misc(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        print('Scraping {} {} miscellaneous stats'.format(year, league))
        season = str(year-1)+'-'+str(year)
        url = self.get_season_link(year,league)
        new = url.split('/')
        new = '/'.join(new[:-1]) + '/misc/' + new[-1]
        new = new.replace("https","http")
        if player:
            self.driver.get(new)
            if normalize:
                self.normalize_table("//*[@id=\"stats_misc_per_match_toggle\"]")
            # get html and scrape table
            html = self.get_html_w_id('stats_misc')
            df = pd.read_html(html)[0]
            # drop duplicate header rows and link to match logs
            df = df[df[("Unnamed: 0_level_0","Rk")]!="Rk"].reset_index(drop=True)
            df.drop(columns="Matches", level=1, inplace=True)
            # convert type from str to float
            for col in list(df.columns.get_level_values(0)):
                if 'Unnamed' not in col:
                    df[col] = df[col].astype("float")
            df = self.add_player_ids_and_links(df, new) # get player IDs
            return df
        else:
            df = pd.read_html(new)
            squad = df[0].copy()
            vs = df[1].copy()
            if normalize:
                if year >= 2018:
                    keep_cols = [("Aerial Duels","Won%")]
                    keep = squad[keep_cols]
                    squad.iloc[:,3:] = squad.iloc[:,3:].divide(squad[("Unnamed: 2_level_0","90s")], axis="rows")
                    squad[keep_cols] = keep
                    keep = vs[keep_cols]
                    vs.iloc[:,3:] = vs.iloc[:,3:].divide(vs[("Unnamed: 2_level_0","90s")], axis="rows")
                    vs[keep_cols] = keep
                else:
                    squad.iloc[:,3:] = squad.iloc[:,3:].divide(squad[("Unnamed: 2_level_0","90s")], axis="rows")
                    vs.iloc[:,3:] = vs.iloc[:,3:].divide(vs[("Unnamed: 2_level_0","90s")], axis="rows")
            # Get team IDs
            squad = self.add_team_ids(squad, 1, new, 'th') 
            vs = self.add_team_ids(vs, 1, new, 'th')
            return squad, vs
        
    ################################################################################
    def scrape_season(self, year, league, normalize=False, player=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        if year >= 2018:
            out = {"League Table":         self.scrape_league_table(year,league,normalize),
                   "Standard":             self.scrape_standard(year,league,normalize,player),
                   "Goalkeeping":          self.scrape_gk(year,league,normalize,player),
                   "Advanced Goalkeeping": self.scrape_adv_gk(year,league,normalize,player),
                   "Shooting":             self.scrape_shooting(year,league,normalize,player),
                   "Passing":              self.scrape_passing(year,league,normalize,player),
                   "Pass Types":           self.scrape_passing_types(year,league,normalize,player),
                   "GCA":                  self.scrape_goal_shot_creation(year,league,normalize,player),
                   "Defensive":            self.scrape_defensive(year,league,normalize,player),
                   "Possession":           self.scrape_possession(year,league,normalize,player),
                   "Playing Time":         self.scrape_playing_time(year,league,normalize,player),
                   "Misc":                 self.scrape_misc(year,league,normalize,player)}
        else:
            out = {"League Table":         self.scrape_league_table(year,league,normalize),
                   "Standard":             self.scrape_standard(year,league,normalize,player),
                   "Goalkeeping":          self.scrape_gk(year,league,normalize,player),
                   "Shooting":             self.scrape_shooting(year,league,normalize,player),
                   "Playing Time":         self.scrape_playing_time(year,league,normalize,player),
                   "Misc":                 self.scrape_misc(year,league,normalize,player)}
        return out

    ################################################################################
    def scrape_matches(self, year, league, save=False):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        season = str(year-1)+'-'+str(year)
        links = self.get_match_links(year,league)
        failures = []
        
        # initialize df
        if year >= 2018:
            cols = ['Date','Matchweek','Home Team','Away Team','Home Goals','Away Goals',
                    'Home Ast','Away Ast','FBRef Home xG','FBRef Away xG','FBRef Home npxG',
                    'FBRef Away npxG','FBRef Home xA','FBRef Away xA','FBRef Home psxG',
                    'FBRef Away psxG', 'Home Player Stats', 'Away Player Stats']
        else:
            cols = ['Date','Matchweek','Home Team','Away Team','Home Goals','Away Goals']
        matches = pd.DataFrame(columns=cols)
        
        # scrape match data
        print("Scraping matches.")
        time.sleep(1)
        for link in tqdm(links):
            try:
                match = self.scrape_match(link, year, league)
                matches = matches.append(match, ignore_index=True)
            except Exception as e:
                failures.append([link, e])
            
        # sort df by match date
        matches = matches.sort_values(by='Date').reset_index(drop=True)
            
        # Print out the failed scrapes
        if len(failures) > 0:
            print('Unable to scrape match data from')
            for failure in failures:
                print(failure, '\n')
        
        # save to CSV if requested by user
        if save:
            filename = '{}_{}_FBRef_matches.csv'.format(season, league.replace(' ','_'))
            matches.to_csv(path_or_buf=filename, index=False)
            print('Matches dataframe saved to ' + filename)
            return filename
        else:
            return matches
        
    ################################################################################    
    def scrape_match(self, link, year, league):
        err, valid = check_season(year,league,'FBRef')
        if not valid:
            print(err)
            return -1
        df = pd.read_html(link)
        
        """
        Earlier than 2017/18 tables
        0) home lineup
        1) away lineup
        2) team stats (possession, pass acc, sot, saves)
        3) home sumary
        4) home gk
        5) away summary
        6) away gk
        
        2017/18 or later tables
        read_html returns 17 dataframes for each match link
        0) home lineup
        1) away lineup
        2) team stats (possession, pass acc, sot, saves)
        3) home summary
        4) home passing
        5) home pass types
        6) home def actions
        7) home possession
        8) home misc
        9) home gk
        10) away summary
        11) away passing
        12) away pass types
        13) away def actions
        14) away possession
        15) away misc
        16) away gk
        17) list of shots - both teams
        18) list of shots - home team
        19) list of shots - away team
        """
        
        if league == 'EPL':
            if year >= 2008:
                spliton = '-Premier-League'
            else:
                spliton = '-Premiership'
        elif league == 'La Liga':
            spliton = '-La-Liga'
        elif league == 'Bundesliga':
            spliton = '-Bundesliga'
        elif league == 'Serie A':
            spliton = '-Serie-A'
        elif league == 'Ligue 1':
            if year >= 2003:
                spliton = '-Ligue-1'
            else:
                spliton = '-Division-1'
        
        # Get date of the match
        try:
            # Try this first. Assumes league name is one word
            date_elements = link.split("/")[-1].split("-")[-4:-1]
            date = '-'.join(date_elements)
            date = datetime.datetime.strptime(date,'%B-%d-%Y').date()
        except:
            # Assumes league name is two words
            date_elements = link.split("/")[-1].split("-")[-5:-2]
            date = '-'.join(date_elements)
            date = datetime.datetime.strptime(date,'%B-%d-%Y').date()
            
        # Get matchweek
        response = requests.get(link)
        soup = BeautifulSoup(response.content, "html.parser")
        dom = etree.HTML(str(soup))
        matchweek = int(
            dom.xpath('//*[@id="content"]/div[2]/div[3]/div[2]/text()')[0]\
                .split('Matchweek')[-1]\
                .replace(')','')\
                .strip()
        )

        match = pd.Series()
        match['Date'] = str(date)
        match['Matchweek'] = matchweek
        match['Home Team'] = df[2].columns[0][0]
        match['Away Team'] = df[2].columns[1][0]
        match["Home Formation"] = df[0].columns[0].split("(")[-1].split(")")[0]
        match["Away Formation"] = df[1].columns[0].split("(")[-1].split(")")[0]
        
        if year >= 2018:
            match['Home Goals'] = np.array(df[3][('Performance','Gls')])[-1]
            match['Away Goals'] = np.array(df[10][('Performance','Gls')])[-1]
            match['Home Ast']   = np.array(df[3][('Performance','Ast')])[-1]
            match['Away Ast']   = np.array(df[10][('Performance','Ast')])[-1]
            
            match['FBRef Home xG']   = np.array(df[3][('Expected','xG')])[-1]
            match['FBRef Away xG']   = np.array(df[10][('Expected','xG')])[-1]
            match['FBRef Home npxG'] = np.array(df[3][('Expected','npxG')])[-1]
            match['FBRef Away npxG'] = np.array(df[10][('Expected','npxG')])[-1]
            match['FBRef Home xA']   = np.array(df[3][('Expected','xA')])[-1]
            match['FBRef Away xA']   = np.array(df[10][('Expected','xA')])[-1]
            match['FBRef Home psxG'] = np.array(df[16][('Shot Stopping','PSxG')])[-1]
            match['FBRef Away psxG'] = np.array(df[9][('Shot Stopping','PSxG')])[-1]
            
            match["Home Player Stats"] = pd.Series({"Team Sheet": df[0],
                                                    "Summary": df[3],
                                                    "Passing": df[4],
                                                    "Pass Types": df[5],
                                                    "Defensive": df[6],
                                                    "Possession": df[7],
                                                    "Misc": df[8],
                                                    "GK": df[9]})
            match["Away Player Stats"] = pd.Series({"Team Sheet": df[1],
                                                    "Summary": df[10],
                                                    "Passing": df[11],
                                                    "Pass Types": df[12],
                                                    "Defensive": df[13],
                                                    "Possession": df[14],
                                                    "Misc": df[15],
                                                    "GK": df[16]})
            # added by @hedonistrh
            match["Shots"] = pd.Series({"Both Team Shots": df[17].loc[~df[17].isna().all(axis=1)], 
                                        "Home Team Shots": df[18].loc[~df[18].isna().all(axis=1)],
                                        "Away Team Shots": df[19].loc[~df[19].isna().all(axis=1)]})
        else:
            match['Home Goals'] = np.array(df[3][('Performance','Gls')])[-1]
            match['Away Goals'] = np.array(df[5][('Performance','Gls')])[-1]
            match['Home Ast']   = np.array(df[3][('Performance','Ast')])[-1]
            match['Away Ast']   = np.array(df[5][('Performance','Ast')])[-1]
            
        return match
    
    ################################################################################
    def scrape_complete_scouting_reports(self, year, league, goalkeepers=False):
        # Get the player links
        if goalkeepers:
            player_links = self.scrape_gk(year, league, player=True)['player_link']
        else:
            player_links = self.scrape_standard(year, league, player=True)['player_link']
        clear_output()
        
        # initialize dataframes
        per90_df = pd.DataFrame()
        percentiles_df = pd.DataFrame()
        
        # gather complete reports and append to dataframes
        cnt = 0
        for player_link in player_links:
            cnt += 1
            print('{}/{}'.format(cnt, len(player_links)), end='\r')
            _, per90, percentiles = self.complete_report_from_player_link(player_link)
            
            # skip goalkeers or players who don't have a complete report
            if (type(per90) is int) or (type(percentiles) is int) \
                    or not goalkeepers and per90['Position'].values[0]=='Goalkeepers':
                continue
                
            # append
            per90_df = per90_df.append(per90, ignore_index=True)
            percentiles_df = percentiles_df.append(percentiles, ignore_index=True)
        
        return per90_df, percentiles_df
    
    ################################################################################
    def complete_report_from_player_link(self, player_link):
        # return -1 if the player has no scouting report
        player_link_html = urlopen(player_link).read().decode('utf8')
        if 'view complete scouting report' not in player_link_html.lower():
            return -1, -1, -1

        # Get the link to the complete report
        self.driver.get(player_link)
        complete_report_button = self.driver.find_element(By.XPATH, 
                                                          '/html/body/div[2]/div[6]/div[2]/div[1]/div/div/div[1]/div/ul/li[2]/a')
        complete_report_link = complete_report_button.get_attribute('href')

        
        self.driver.get(complete_report_link)

        # Get the report table
        complete_report = pd.read_html(complete_report_link)[0]
        complete_report.columns = complete_report.columns.get_level_values(1)
        complete_report = pd.concat([pd.DataFrame(data={'Statistic': ['Standard Stats','Statistic'], 
                                                        'Per 90': ['Standard Stats','Per 90'], 
                                                        'Percentile': ['Standard Stats', 'Percentile']}),
                                     complete_report])
        complete_report.dropna(axis=0, inplace=True)
        complete_report.reset_index(inplace=True, drop=True)

        # Get the table section headers and stats to make a multiindex
        header_idxs = [i for i in complete_report.index \
                       if np.all(complete_report.iloc[i,:]==complete_report.iloc[i,0])]
        header_idxs.append(complete_report.shape[0])
        table_headers = list()
        sub_stats = list()
        for i in range(len(header_idxs)-1):
            table_headers.append(complete_report.iloc[header_idxs[i],0])
            table = complete_report.iloc[header_idxs[i]+2:header_idxs[i+1], :].T
            sub_stats.append(list(table.iloc[0,:].values)) # sub stats are in first row
        idx = pd.MultiIndex.from_tuples([(table_headers[i], sub_stats[i][j]) \
                                         for i in range(len(table_headers)) \
                                         for j in range(len(sub_stats[i]))])
        
        # Initiate the dataframes
        per90 = pd.DataFrame(data=-1*np.ones([1,len(idx)]), columns=idx)
        percentiles = pd.DataFrame(data=-1*np.ones([1,len(idx)]), columns=idx)
        
        # Populate the dataframes
        for i in range(len(header_idxs)-1):
            table_header = complete_report.iloc[header_idxs[i],0]
            table = complete_report.iloc[header_idxs[i]+2:header_idxs[i+1], :].T
            table.columns = table.iloc[0,:]
            table = table.reset_index(drop=True).drop(index=0)
            for col in table.columns:
                per90[(table_header,col)] = float(table.loc[1,col].replace('%',''))
                percentiles[(table_header,col)] = int(table.loc[2,col])
        
        # add player names, positions, and minutes played
        player_name = ' '.join(complete_report_link.split('/')[-1].split('-')[:-2])
        player_pos = self.driver.find_element(By.XPATH, '//*[@class="filter switcher"]/div/a').text.replace('vs. ', '')
        minutes = int(self.driver.find_element(By.XPATH, '//*[@class="footer no_hide_long"]/div') \
                .text \
                .split(' minutes')[0] \
                .split(' ')[-1])
        per90['Player'] = player_name
        per90['Position'] = player_pos
        per90['Minutes'] = minutes
        percentiles['Player'] = player_name
        percentiles['Position'] = player_pos
        percentiles['Minutes'] = minutes

        return complete_report, per90, percentiles
        


from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from webdriver_manager.chrome import ChromeDriverManager
import os
import pandas as pd
import numpy as np
from IPython.display import clear_output
from zipfile import ZipFile
from sportstat.shared_functions import check_season

class FiveThirtyEight:
    
    def __init__(self):
        options = Options()
        options.headless = True
        prefs = {"download.default_directory" : os.getcwd()}
        options.add_experimental_option("prefs",prefs)
        self.driver = webdriver.Chrome(ChromeDriverManager().install(), options=options)
        clear_output()
        
        
    def close(self):
        self.driver.close()
        self.driver.quit()
        
        
    def up_season(self,string):
        integer = int(string)
        up1 = integer + 1
        return str(up1)
        
        
    def scrape_matches(self, year, league, save=False):
        if not check_season(year,'EPL','FiveThirtyEight'):
            return -1
        
        url = 'https://data.fivethirtyeight.com/#soccer-spi'
        self.driver.get(url)
        
        # find and click download button
        for element in self.driver.find_elements_by_class_name("download-link"):
            if element.get_attribute("dataset-name") == "soccer-spi":
                element.click()
                
        # wait for download to complete
        while not os.path.exists('soccer-spi.zip'):
            pass
        
        # get data table
        with ZipFile('soccer-spi.zip') as zf:
            with zf.open('soccer-spi/spi_matches.csv') as f:
                df = pd.read_csv(f)
        os.remove('soccer-spi.zip') # delete downloaded folder
        
        # pick the chosen league
        if league == "EPL":
            df = df[df['league'] == 'Barclays Premier League']
        elif league == "La Liga":
            df = df[df['league'] == 'Spanish Primera Division']
        elif league == "Bundesliga":
            df = df[df['league'] == 'German Bundesliga']
        elif league == "Serie A":
            df = df[df['league'] == 'Italy Serie A']
        elif league == "Ligue 1":
            df = df[df['league'] == 'French Ligue 1']
        
        # add one to season column
        df['season'] = df['season'].apply(self.up_season)
        
        # only keep the season requested
        df = df[df['season']==str(year)]
        
        # only keep some cols
        keep = ['date','team1','team2','score1','score2','xg1','xg2',
                'nsxg1','nsxg2','adj_score1','adj_score2']
        df = df[keep]
        
        # rename cols
        cols = ['Date','Home Team','Away Team','Home Goals','Away Goals',
                'FTE Home xG','FTE Away xG','FTE Home nsxG','FTE Away nsxG',
                'Home Adj Score','Away Adj Score']
        df.columns = cols
        
        df = df.reset_index(drop=True)
        
        # save to CSV if requested by user
        if save:
            filename = season+"_"+league+"_FiveThirtyEight_matches.csv"
            df.to_csv(path_or_buf=filename, index=False)
            print('Matches dataframe saved to ' + filename)
            return filename
        else:
            return df
        


from sportstat.FBRef import FBRef
from sportstat.FiveThirtyEight import FiveThirtyEight
from sportstat.sportstat import ScraperFC
from sportstat.Understat import Understat
from sportstat.SofaScore import SofaScore
from sportstat.ClubElo import ClubElo
from sportstat.Capology import Capology
from sportstat.Transfermarkt import Transfermarkt
from sportstat.shared_functions import *


from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from webdriver_manager.chrome import ChromeDriverManager
from sportstat.shared_functions import *
from IPython.display import clear_output
import time

class SofaScore:

    def __init__(self):
        options = Options()
        # options.headless = True
        options.add_argument("window-size=1400,600")
        self.driver = webdriver.Chrome(ChromeDriverManager().install(), options=options)
        clear_output()


    def close(self):
        self.driver.close()
        self.driver.quit()


    def scrape_team_stats(self, year, league, normalize=True):
        error, valid = check_season(year, league, "SofaScore")
        if not valid:
            print(error)
            return -1
        
        url = "https://www.sofascore.com/tournament/football/usa/usl-league-one/13362"
        self.driver.get(url)

        # Click button for right season
        for el in self.driver.find_elements_by_tag_name("li"):
            if str(year) in el.get_attribute("outerHTML"):
                button = el
                break
        self.driver.execute_script("arguments[0].click()",button)
        time.sleep(3)

        # Get links to the teams in the tables
        team_links = list()
        for el in self.driver.find_elements_by_tag_name("a"):
            href = el.get_attribute("href")
            if href and ("team" in href):
                team_links.append(href)

        # Scrape stats for each team
        for team_link in team_links[:1]:
            self.driver.get(team_link) # Go to team's page

            # Click to season stats for the chosen season
            buttons = list()
            for el in self.driver.find_elements_by_tag_name("li"):
                time.sleep(0.1)
                if str(year) in el.get_attribute("outerHTML"):
                    buttons.append(el)
            self.driver.execute_script("arguments[0].click()",buttons[1])

            matches = None
            team_stats = {
                "Team name": team_link.split("/")[-2].replace("-"," "),
                "Avg. rating": self.driver.find_element_by_xpath("/html/body/div[1]/main/div/div[2]/div[1]/div[2]/div/div[4]/div/div[2]/div[2]").text,
                "Matches": matches,
                "GF": None,
                "GA": None,
                "A": None,
                "Goal conversion": None,
                "PK goals": None,
                "FK goals": None,
                "Goals from inside the box": None,
                "Goals from outside the box": None,
                "Left foot goals": None,
                "Right foot goals": None,
                "Headed goals": None,
                "Big chances": round(None*matches),
                "Big chances missed": round(None*matches),
                "Shots": round(None*matches),
                "Shots on target": round(None*matches),
                "Shots off target": round(None*matches),
                "Succ. dribbles": round(None*matches),
                "Corners": round(None*matches),
                "Hit woodwork": None,
                "Counter attacks": None,
                "Avg. possession": None,
                "Acc. passes": round(None*matches),
                "Pass acc.": None,
                "Comp. passes own half": None,
                "Pass acc. own half": None,
                "Comp. passes opp. half": None,
                "Pass acc. opp. half": None,
                "Comp. long balls": None,
                "Long ball acc.": None,
                "Comp. crosses": None,
                "Cross acc.": None,
                "CS": None,
                "Tackles": round(None*matches),
                "Interceptions": round(None*matches),
                "Clearances": round(None*matches),
                "Saves": round(None*matches),
                "Err. leading to shot": None,
                "Err leading to goal": None,
                "PK conceded": None,
                "PK goals conceded": None,
                "Clearances off line": None,
                "Last man tackles": None,
                "Duels won": round(None*matches),
                "Duels win %": None,
                "Ground duels won": round(None*matches),
                "Ground duels win %": None,
                "Aerial duels won": round(None*matches),
                "Aerial duels win %": None,
                "Poss. lost": round(None*matches),
                "Offsides": round(None*matches),
                "Fouls": round(None*matches),
                "YC": round(None*matches),
                "RC": None,
                "Squad size": None
            }

        if normalize:
            pass

        return 999




import datetime
import json
import numpy as np
import pandas as pd
from sportstat.shared_functions import check_season
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from webdriver_manager.chrome import ChromeDriverManager
from tqdm import tqdm
import requests


class Understat:
    
    ############################################################################
    def __init__(self):
        options = Options()
        options.headless = True
        options.add_argument("window-size=1400,600")
        prefs = {'profile.managed_default_content_settings.images': 2} # don't load images
        options.add_experimental_option('prefs', prefs)
        self.driver = webdriver.Chrome(ChromeDriverManager().install(), options=options)
        
        
    ############################################################################   
    def close(self):
        self.driver.close()
        self.driver.quit()
        
        
    ############################################################################    
    def get_season_link(self, year, league):
        """
        Gets URL of the season.
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :return: string, url of the year and league passed in
        """
        lg = league.replace(" ","_")
        url = f"https://understat.com/league/{lg}/{str(year-1)}"
        return url
        
        
    ############################################################################    
    def get_match_links(self, year, league):
        """
        Gets all of the match links for the league and year
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :return: set, match links
        """
        if not check_season(year,league,'Understat'):
            return -1
        base_url = "https://understat.com/"
        lg = league.replace(" ","_")
        url = base_url+"league/"+lg+"/"+str(year-1)
        self.driver.get(url)
        
        # Gather the links by hitting the "prev week" button until it's disabled
        links = set()
        btn = self.driver.find_element(By.CLASS_NAME, "calendar-prev")

        while "disabled" not in btn.get_attribute("outerHTML"):
            for el in self.driver.find_elements(By.CLASS_NAME, "match-info"):
                links.add(el.get_attribute("href"))
            btn.click()
            
        # One final collection for first week of season
        for el in self.driver.find_elements(By.CLASS_NAME, "match-info"):
            links.add(el.get_attribute("href"))
                
        return links
    
    ############################################################################
    def get_team_links(self, year, league):
        """
        Gets all of the team links for the league and year
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :return: set, links to teams' pages for the year
        """
        team_links = set()
        self.driver.get(self.get_season_link(year, league))
        for el in self.driver.find_elements(By.TAG_NAME, "a"):
            href = el.get_attribute("href")
            if href and "team" in href:
                team_links.add(href)
        return list(team_links)
    
    
    ############################################################################
    def remove_diff(self, string):
        """
        Removes the plus/minus from some stats like xG
        :param string: the string to remove the difference from
        :return: the string with the difference removed
        """
        string = string.split("-")[0]
        return float(string.split("+")[0])
    
    
    ############################################################################
    def unhide_stats(self, columns):
        """
        Understat doesn't display all stats by default. Uses the stats currently 
        shown in the table columns to unhide stats that aren't being displayed.
        :param columns: the columns currently shown in the table being scraped
        :return: None
        """
        # Show the options popup
        self.driver.find_elements(By.CLASS_NAME, "options-button")[0].click()
        
        # Iterate across all stats and show the ones that aren't in columns
        for el in self.driver.find_elements(By.CLASS_NAME, "table-options-row"):
            stat_name = el.find_element(By.CLASS_NAME, "row-title").text
            if (stat_name not in columns) and (stat_name != ""):
                el.find_element(By.CLASS_NAME, "row-display").click()

        # click the apply button
        self.driver.find_elements(By.CLASS_NAME, "button-apply")[0].click()
        
        return
        
        
    ############################################################################   
    def scrape_match(self, link):
        """
        Scrapes a single match from Understat.
        :param link: URL to the match
        :return: Pandas series of match stats
        """
        self.driver.get(link)
        
        elements = []
        for element in self.driver.find_elements(By.CLASS_NAME, "progress-value"):
            elements.append(element.get_attribute("innerHTML"))

        match = pd.Series()
        
        """ Match date and ID """
        for element in self.driver.find_elements(By.CLASS_NAME, "breadcrumb"):
            date = element.find_elements(By.TAG_NAME, "li")[2]
            date = date.text
        date = datetime.datetime.strptime(date,'%b %d %Y').date()
        match['date'] = date
        match['match id'] = link.split("/")[-1]


        """ Team Names """
        progress_values = [el.get_attribute("innerHTML") \
                           for el in self.driver.find_elements(By.CLASS_NAME, "progress-value")]
        match['home team'] = progress_values[0]
        match['away team'] = progress_values[1]


        """ Data from team stats table """
        # Go to team stats table
        [el for el in self.driver.find_elements(By.TAG_NAME, "label") \
         if "Stats" in el.text][0].click()
        # Each row in the stats table is a progress bar
        prog_bars = self.driver.find_elements(By.CLASS_NAME, "progress-bar")

        # Chances (separated out cuz it utilizes a draw element)
        bar = prog_bars[1]
        home = bar.find_element(By.CLASS_NAME, "progress-home").text
        try:
            draw = bar.find_element(By.CLASS_NAME, "progress-draw").text
        except NoSuchElementException:
            draw = "0%"
        away = bar.find_element(By.CLASS_NAME, "progress-away").text

        match["home win probability"] = float(home.replace("%","")) if home!="" else 0
        match["draw probability"]     = float(draw.replace("%","")) if draw!="" else 0
        match["away win probability"] = float(away.replace("%","")) if away!="" else 0

        # Goals, xG, shots, DEEP, PPDA, xPTS
        for bar in prog_bars[2:]:
            stat = bar.find_element(By.CLASS_NAME, "progress-title").text.lower()
            home = bar.find_element(By.CLASS_NAME, "progress-home").text
            away = bar.find_element(By.CLASS_NAME, "progress-away").text
            match[f"home {stat}"] = float(home)
            match[f"away {stat}"] = float(away)


        """ Data from player stats tables """
        #### Show all hidden stats ####
        columns = pd.read_html(self.driver.find_element(By.TAG_NAME, "table")\
                                          .get_attribute("outerHTML"))[0].columns
        self.unhide_stats(columns)

        #### Scrape player stats tables ####
        home_player_df = pd.read_html(self.driver.find_element(By.TAG_NAME, "table")\
                                                 .get_attribute("outerHTML"))[0]
        # Click button to go to away player stats table
        [el for el in self.driver.find_elements(By.TAG_NAME, "label") \
         if el.get_attribute("for")=="team-away"][0].click()
        away_player_df = pd.read_html(self.driver.find_element(By.TAG_NAME, "table")\
                                                 .get_attribute("outerHTML"))[0]

        #### Get the team sum stats that weren't gathered in the team stats table ####
        for team, temp in (("home", home_player_df), ("away", away_player_df)):
            for stat, column in (("key passes","KP"), 
                                 ("xa","xA"), 
                                 ("xgchain","xGChain"), 
                                 ("xgbuildup","xGBuildup")):
                match[f"{team} {stat}"] = self.remove_diff(str(temp.loc[temp.index[-1], column]))

        #### Remove No column, remove sum row, and remove diffs on xG and xA columns ####
        for player_df in (home_player_df, away_player_df):
            # drop sum row and No columns
            player_df.drop(index=player_df.index[-1], columns=player_df.columns[0], inplace=True) 
            # remove diffs
            player_df[["xG","xA"]] = player_df[["xG","xA"]].astype(str) # need to be str for remove_diff()
            player_df["xG"] = player_df.apply(lambda row : self.remove_diff(row["xG"]), axis=1)
            player_df["xA"] = player_df.apply(lambda row : self.remove_diff(row["xA"]), axis=1)

        #### Player stats dfs to match series ####
        match["home player stats"] = home_player_df
        match["away player stats"] = away_player_df
        
        return match
        
        
    ############################################################################    
    def scrape_matches(self, year, league, save=False):
        """
        Scrapes all of the matches from the passed year and league
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :param save: boolean, if true the output is saved to a CSV
        :return: Pandas DataFrame if not saved, filepath to CSV if saved
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        season = str(year-1)+'-'+str(year)
        links = self.get_match_links(year, league)
        matches = pd.DataFrame()
        
        for link in tqdm(links):
#             print(link)
            match   = self.scrape_match(link)
            matches = matches.append(match, ignore_index=True)
        
        # save to CSV if requested by user
        if save:
            filename = season+"_"+league+"_Understat_matches.csv"
            matches.to_csv(path_or_buf=filename, index=False)
            print('Matches dataframe saved to ' + filename)
            return filename
        else:
            return matches
        
        
    ############################################################################    
    def scrape_league_table(self, year, league, normalize=False):
        """
        Scrapes the league table from the year passed in
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :param normalize: boolean, if true the output is normalized by the number
                          of matches played
        :return: Pandas DataFrame of the league table
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        url = self.get_season_link(year, league) # link to the selected league/season
        self.driver.get(url)
        
        """ Show all of the stats """
        # Get column names currently show in the table
        table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute("outerHTML")
        columns = pd.read_html(table)[0].columns
        self.unhide_stats(columns)

        """ dataframe """
        table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute("outerHTML")
        df = pd.read_html(table)[0]
        
        # remove performance differential text from some columns
        for i in range(df.shape[0]):
            df.loc[i,"xG"] = self.remove_diff(df.loc[i,"xG"])
            df.loc[i,"xGA"] = self.remove_diff(df.loc[i,"xGA"])
            df.loc[i,"xPTS"] = self.remove_diff(df.loc[i,"xPTS"])
            
        if normalize:
            df.iloc[:,3:14] = df.iloc[:,3:14].divide(df["M"], axis="rows")
            df.iloc[:,16:] = df.iloc[:,16:].divide(df["M"], axis="rows")
        
        self.close()
        self.__init__()
        return df
    
    
    ############################################################################
    def scrape_home_away_tables(self, year, league, normalize=False):
        """
        Scrapes the home and away league tables from the year passed in
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :param normalize: boolean, if true the output is normalized by the number
                          of matches played
        :return: home - the home league table, Pandas DataFrame
                 away - the away league table, Pandas DataFrame
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        url = self.get_season_link(year, league) # link to the selected league/season
        self.driver.get(url)

        """ Show all of the stats """
        # Get columns that are already shown
        labels = self.driver.find_elements(By.TAG_NAME, "label")
        [el for el in labels if el.text=='home'][0].click()
        table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute('outerHTML')
        columns = pd.read_html(table)[0].columns 
        self.unhide_stats(columns)
        
        """ Home Table """
        labels = self.driver.find_elements(By.TAG_NAME, "label")
        [el for el in labels if el.text=='home'][0].click()
        table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute('outerHTML')
        home = pd.read_html(table)[0]
        
        
        """ Away Table """
        [el for el in labels if el.text=='away'][0].click()
        table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute('outerHTML')
        away = pd.read_html(table)[0]
        
        # remove differentials from some columns
        for i in range(home.shape[0]):
            home.loc[i,"xG"] = self.remove_diff(home.loc[i,"xG"])
            home.loc[i,"xGA"] = self.remove_diff(home.loc[i,"xGA"])
            home.loc[i,"xPTS"] = self.remove_diff(home.loc[i,"xPTS"])
            away.loc[i,"xG"] = self.remove_diff(away.loc[i,"xG"])
            away.loc[i,"xGA"] = self.remove_diff(away.loc[i,"xGA"])
            away.loc[i,"xPTS"] = self.remove_diff(away.loc[i,"xPTS"])
        
        if normalize:
            home.iloc[:,3:14] = home.iloc[:,3:14].divide(home["M"], axis="rows")
            home.iloc[:,16:] = home.iloc[:,16:].divide(home["M"], axis="rows")
            away.iloc[:,3:14] = away.iloc[:,3:14].divide(away["M"], axis="rows")
            away.iloc[:,16:] = away.iloc[:,16:].divide(away["M"], axis="rows")
        
        self.close()
        self.__init__()
        return home, away
    
    
    ############################################################################
    def scrape_situations(self, year, league):
        """
        Scrapes the situations leading to shots for each team in the league and year
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :return: Pandas DataFrame
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        # Get links for teams in league that season
        team_links = self.get_team_links(year, league)
                
        mi = pd.MultiIndex.from_product(
            [["Open play", "From corner", "Set piece", "Direct FK", "Penalty"],
             ["Sh", "G", "ShA", "GA", "xG", "xGA", "xGD", "xG/Sh", "xGA/Sh"]]
        )
        mi = mi.insert(0, ("Team names", "Team"))
        situations = pd.DataFrame()#columns=mi)
        
        for link in team_links:
            
            team_name = link.split("/")[-2]
            self.driver.get(link)
            table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute("outerHTML")
            df = pd.read_html(table)[0]
            
            for i in range(df.shape[0]):
                # remove performance differential text from some columns
                df.loc[i,"xG"] = self.remove_diff(df.loc[i,"xG"])
                df.loc[i,"xGA"] = self.remove_diff(df.loc[i,"xGA"])
            
            # reformat df to fit into a row
            df.drop(columns=["№","Situation"], inplace=True)
            row = df.to_numpy()
            row = np.insert(row, 0, team_name) # insert team name
            
            # append row
            situations = situations.append(
                pd.DataFrame(row.reshape(1,-1)),#, columns=situations.columns),
                ignore_index=True
            )
        
        situations.columns = mi
        self.close()
        self.__init__()
        return situations
    
    
    ############################################################################
    def scrape_formations(self, year, league):
        """
        Scrapes the stats for each team in the year and league, broken down by 
        formation used by the team.
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :return: dict, each team is a key and each team has their own dict for
                 each formation used
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        # Get links for teams in league that season
        team_links = self.get_team_links(year, league)
        
        formations = dict()
        
        for link in team_links:
            
            # Get team name to add to formations
            team_name = link.split("/")[-2]
            if team_name not in formations.keys():
                formations[team_name] = dict()
                
            # Got to team's link, click formations, and get table of formations used
            self.driver.get(link)
            self.driver.find_element(
                By.XPATH, 
                "/html/body/div[1]/div[3]/div[3]/div/div[1]/div/label[2]"
            ).click()
            
            table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute("outerHTML")
            df = pd.read_html(table)[0]
            df.drop(columns=["№"], inplace=True)
            
            # Remove performance differential text from some columns
            for i in range(df.shape[0]):
                df.loc[i,"xG"] = self.remove_diff(df.loc[i,"xG"])
                df.loc[i,"xGA"] = self.remove_diff(df.loc[i,"xGA"])
                
                formation = df.loc[i,"Formation"]
                
                if formation not in formations[team_name].keys():
                    formations[team_name][formation] = df.iloc[i,:].drop(columns=["№","Formation"])                
                
        self.close()
        self.__init__()
        return formations
    
    
    ############################################################################
    def scrape_game_states(self, year, league):
        """
        Scrapes the game states for each team in the year and league
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :return: Pandas DataFrame
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        # Get links for teams in league that season
        team_links = self.get_team_links(year, league)
                
        mi = pd.MultiIndex.from_product(
            [["Goal diff 0", "Goal diff -1", "Goal diff +1", "Goal diff < -1", "Goal diff > +1"],
             ["Min", "Sh", "G", "ShA", "GA", "xG", "xGA", "xGD", "xG90", "xGA90"]]
        )
        mi = mi.insert(0, ("Team names", "Team"))
        game_states = pd.DataFrame()#columns=mi)
        
        for link in team_links:
            
            team_name = link.split("/")[-2]
            self.driver.get(link)
            self.driver.find_element(
                By.XPATH, 
                "/html/body/div[1]/div[3]/div[3]/div/div[1]/div/label[3]"
            ).click()
            table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute("outerHTML")
            df = pd.read_html(table)[0]
            df.drop(columns=["№"], inplace=True)
            
            row = {"Goal diff 0": None,
                   "Goal diff -1": None,
                   "Goal diff +1": None,
                   "Goal diff < -1": None,
                   "Goal diff > +1": None}
            for i in range(df.shape[0]):
                # remove performance differential text from some columns
                df.loc[i,"xG"] = self.remove_diff(df.loc[i,"xG"])
                df.loc[i,"xGA"] = self.remove_diff(df.loc[i,"xGA"])
                
                game_state = df.loc[i,"Game state"]
                row[game_state] = df.loc[i,:].drop(labels=["Game state"])
                
            row_array = []
            for key in row.keys():
                row_array.append( row[key].to_numpy() )
            row_array = np.array(row_array)
            row_array = np.insert(row_array, 0, team_name) # insert team name
            
            # append row
            game_states = game_states.append(
                pd.DataFrame(row_array.reshape(1,-1)),#, columns=game_states.columns),
                ignore_index=True
            )
            
        game_states.columns = mi
        self.close()
        self.__init__()
        return game_states
    
    
    ############################################################################
    def scrape_timing(self, year, league):
        """
        Scrapes the timing of goals for each team in the year and league
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :return: Pandas DataFrame
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        # Get links for teams in league that season
        team_links = self.get_team_links(year, league)

        mi = pd.MultiIndex.from_product(
            [["1-15", "16-30", "31-45", "46-60", "61-75", "76+"],
             ["Sh", "G", "ShA", "GA", "xG", "xGA", "xGD", "xG/Sh", "xGA/Sh"]]
        )
        mi = mi.insert(0, ("Team names", "Team"))
        timing_df = pd.DataFrame()#columns=mi)

        for link in team_links:
            
            team_name = link.split("/")[-2]
            self.driver.get(link)
            self.driver.find_element(
                By.XPATH, 
                "/html/body/div[1]/div[3]/div[3]/div/div[1]/div/label[4]"
            ).click()
            table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute("outerHTML")
            df = pd.read_html(table)[0]
            df.drop(columns=["№"], inplace=True)

            row = {"1-15": None,
                   "16-30": None,
                   "31-45": None,
                   "46-60": None,
                   "61-75": None,
                   "76+": None}
            for i in range(df.shape[0]):
                # remove performance differential text from some columns
                df.loc[i,"xG"] = self.remove_diff(df.loc[i,"xG"])
                df.loc[i,"xGA"] = self.remove_diff(df.loc[i,"xGA"])

                timing = df.loc[i, "Timing"]
                row[timing] = df.loc[i,:].drop(labels=["Timing"])

            row_array = []
            for key in row.keys():
                row_array.append(row[key].to_numpy())
            row_array = np.array(row_array)
            row_array = np.insert(row_array, 0, team_name) # insert team name

            # append row
            timing_df = timing_df.append(
                pd.DataFrame(row_array.reshape(1,-1)),#, columns=timing_df.columns),
                ignore_index=True
            )
            
        timing_df.columns = mi
        self.close()
        self.__init__()
        return timing_df
    
    
    ############################################################################
    def scrape_shot_zones(self, year, league):
        """
        Scrapes the shot zones for each team in the year and league
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :return: Pandas DataFrame
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        # Get links for teams in league that season
        team_links = self.get_team_links(year, league)

        mi = pd.MultiIndex.from_product(
            [["Own goals", "Out of box", "Penalty area", "Six-yard box"],
             ["Sh", "G", "ShA", "GA", "xG", "xGA", "xGD", "xG/Sh", "xGA/Sh"]]
        )
        mi = mi.insert(0, ("Team names", "Team"))
        shot_zones_df = pd.DataFrame()#columns=mi)

        for link in team_links:
            
            team_name = link.split("/")[-2]
            self.driver.get(link)
            self.driver.find_element(
                By.XPATH, 
                "/html/body/div[1]/div[3]/div[3]/div/div[1]/div/label[5]"
            ).click()
            table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute("outerHTML")
            df = pd.read_html(table)[0]
            df.drop(columns=["№"], inplace=True)

            row = {"Own goals": None,
                   "Out of box": None,
                   "Penalty area": None,
                   "Six-yard box": None}
            for i in range(df.shape[0]):
                # remove performance differential text from some columns
                df.loc[i,"xG"] = self.remove_diff(df.loc[i,"xG"])
                df.loc[i,"xGA"] = self.remove_diff(df.loc[i,"xGA"])

                zone = df.loc[i, "Shot zones"]
                row[zone] = df.loc[i,:].drop(labels=["Shot zones"])

            row_array = []
            for key in row.keys():
                if row[key] is None:
                    row[key] = pd.Series(np.zeros((9)))
                row_array.append(row[key].to_numpy())
            row_array = np.array(row_array)
            row_array = np.insert(row_array, 0, team_name) # insert team name

            # append row
            shot_zones_df = shot_zones_df.append(
                pd.DataFrame(row_array.reshape(1,-1)),#, columns=shot_zones_df.columns),
                ignore_index=True
            )
            
        shot_zones_df.columns = mi
        self.close()
        self.__init__()
        return shot_zones_df
    
    
    ############################################################################
    def scrape_attack_speeds(self, year, league):
        """
        Scrapes the attack speeds for each team in the year and league
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :return: Pandas DataFrame
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        # Get links for teams in league that season
        team_links = self.get_team_links(year, league)

        mi = pd.MultiIndex.from_product(
            [["Normal", "Standard", "Slow", "Fast"],
             ["Sh", "G", "ShA", "GA", "xG", "xGA", "xGD", "xG/Sh", "xGA/Sh"]]
        )
        mi = mi.insert(0, ("Team names", "Team"))
        attack_speeds_df = pd.DataFrame()#columns=mi)

        for link in team_links:
            
            team_name = link.split("/")[-2]
            self.driver.get(link)
            self.driver.find_element(
                By.XPATH, 
                 "/html/body/div[1]/div[3]/div[3]/div/div[1]/div/label[6]"
                ).click()
            
            table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute("outerHTML")
            df = pd.read_html(table)[0]
            df.drop(columns=["№"], inplace=True)

            row = {"Normal": None,
                   "Standard": None,
                   "Slow": None,
                   "Fast": None}
            for i in range(df.shape[0]):
                # remove performance differential text from some columns
                df.loc[i,"xG"] = self.remove_diff(df.loc[i,"xG"])
                df.loc[i,"xGA"] = self.remove_diff(df.loc[i,"xGA"])

                speed = df.loc[i, "Attack speed"]
                row[speed] = df.loc[i,:].drop(labels=["Attack speed"])

            row_array = []
            for key in row.keys():
                row_array.append(row[key].to_numpy())
            row_array = np.array(row_array)
            row_array = np.insert(row_array, 0, team_name) # insert team name

            # append row
            attack_speeds_df = attack_speeds_df.append(
                pd.DataFrame(row_array.reshape(1,-1)),#, columns=attack_speeds_df.columns),
                ignore_index=True
            )
            
        attack_speeds_df.columns = mi
        self.close()
        self.__init__()
        return attack_speeds_df
    
    
    ############################################################################
    def scrape_shot_results(self, year, league):
        """
        Scrapes the shot results for each team in the year and league
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :return: Pandas DataFrame
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        # Get links for teams in league that season
        team_links = self.get_team_links(year, league)

        mi = pd.MultiIndex.from_product(
            [["Missed shot", "Goal", "Saved shot", "Blocked shot", "Shot on post"],
             ["Sh", "G", "ShA", "GA", "xG", "xGA", "xGD", "xG/Sh", "xGA/Sh"]]
        )
        mi = mi.insert(0, ("Team names", "Team"))
        shot_results_df = pd.DataFrame()#columns=mi)

        for link in team_links:
            
            team_name = link.split("/")[-2]
            self.driver.get(link)
            self.driver.find_element(By.XPATH, "/html/body/div[1]/div[3]/div[3]/div/div[1]/div/label[7]").click()
            table = self.driver.find_elements(By.TAG_NAME, "table")[0].get_attribute("outerHTML")
            df = pd.read_html(table)[0]
            df.drop(columns=["№"], inplace=True)

            row = {
                "Missed shot": None,
                "Goal": None,
                "Saved shot": None,
                "Blocked shot": None,
                "Shot on post": None
            }
            for i in range(df.shape[0]):
                # remove performance differential text from some columns
                df.loc[i,"xG"] = self.remove_diff(df.loc[i,"xG"])
                df.loc[i,"xGA"] = self.remove_diff(df.loc[i,"xGA"])

                result = df.loc[i, "Result"]
                row[result] = df.loc[i,:].drop(labels=["Result"])

            row_array = []
            for key in row.keys():
                row_array.append(row[key].to_numpy())
            row_array = np.array(row_array)
            row_array = np.insert(row_array, 0, team_name) # insert team name

            # append row
            shot_results_df = shot_results_df.append(
                pd.DataFrame(row_array.reshape(1,-1)),
                # columns=shot_results_df.columns),
                ignore_index=True
            )
            
        shot_results_df.columns = mi
        self.close()
        self.__init__()
        return shot_results_df
            
        
    ############################################################################
    def scrape_shot_xy(self, year, league, save=False):
        """
        Scrapes the info for every shot in the league and year.
        :param year: int, calendary year the season ends in
        :param league: string, the league to be scraped
        :param save: boolean, if True output is saved to a JSON file
        :return: Pandas DataFrame if not saved, filepath to JSON if saved
        """
        if not check_season(year,league,'Understat'):
            return -1
        
        season = str(year-1)+'-'+str(year)
        links = self.get_match_links(year, league)
        shots_data = dict()
        failures = list()
        
        sum_time_selenium = 0
        sum_time_requests = 0

        for link in tqdm(links):
            
            match_id = link.split("/")[-1]
            try:
                game_shots_data = json.loads(
                    requests.get(link).text\
                        .split("shotsData")[1]\
                        .split("JSON.parse(\'")[1]\
                        .split("\')")[0]\
                        .encode("latin-1")\
                        .decode("unicode-escape")
                )
                
                shots_data[match_id] = game_shots_data
            except:
                failures.append(match_id)
                shots_data[match_id] = "Error scraping"
            
        self.close()
        self.__init__()
        
        # print any matches that scraping failed for
        print(f"Failed scraping the following matches: {failures}.")

        # save to JSON file
        if save:
            filename = f"{season}_{league}_shot_xy.json"
            with open(filename, "w") as f:
                f.write(json.dumps(shots_data))
            print(f'Scraping saved to {filename}')
            return filename
        else:
            return shots_data



from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from webdriver_manager.chrome import ChromeDriverManager
from IPython.display import clear_output
import random
import pandas as pd
import numpy as np


################################################################################
def check_season(year, league, source):
    valid = {
        'All': {},
        'FBRef': {
            'EPL': 1993,
            'La Liga': 1989,
            'Bundesliga': 1989,
            'Serie A': 1989,
            'Ligue 1': 1996,
            'MLS': 1996
        },
        'Understat': {
            'EPL': 2015,
            'La Liga': 2015,
            'Bundesliga': 2015,
            'Serie A': 2015,
            'Ligue 1': 2015
        },
        'FiveThirtyEight': {
            'EPL': 2017,
            'La Liga': 2017,
            'Bundesliga': 2017,
            'Serie A': 2017,
            'Ligue 1': 2017
        },
        'SofaScore': {
            'USL League One': 2019
        },
        'Capology': {
            'Bundesliga': 2014,
            '2.Bundesliga': 2020,
            'EPL': 2014,
            'EFL Championship': 2014,
            'Serie A': 2010,
            'Serie B': 2020,
            'La Liga': 2014,
            'La Liga 2': 2020,
            'Ligue 1': 2014,
            'Ligue 2': 2020,
            'Eredivisie': 2014,
            'Primeira Liga': 2014,
            'Scottish PL': 2020,
            'Super Lig': 2014,
            'Belgian 1st Division': 2014
        },
        'Transfermarkt': {
            'EPL': 1993,
            'EFL Championship': 2005,
            'EFL1': 2005,
            'EFL2': 2005,
            'Bundesliga': 1964,
            '2.Bundesliga': 1982,
            'Serie A': 1930,
            'Serie B': 1930,
            'La Liga': 1929,
            'La Liga 2': 1929,
            'Ligue 1': 1970,
            'Ligue 2': 1993,
            'Eredivisie': 1955,
            'Scottish PL': 2004,
            'Super Lig': 1960,
            'Jupiler Pro League': 1987,
            'Liga Nos': 1994,
            'Russian Premier League': 2011,
            'Brasileirao': 2001,
            'Argentina Liga Profesional': 2015,
            'MLS': 1996,
        },
    }
    
    assert source in list(valid.keys())
    error = None
    
    # make sure year is an int
    if type(year) != int:
        error = "Year needs to be an integer."
        return error, False
    
    # Make sure league is a valid string for the source
    if type(league)!=str or league not in list(valid[source].keys()):
        error = 'League must be a string. Options are {}'.format(list(valid[source].keys()))
        return error, False
    
    # Make sure the source has data from the requested league and year
    if year < valid[source][league]:
        error = 'Year invalid for source {} and league {}. Must be {} or later.'.format(source, league, valid[source][league])
        return error, False
    
    return error, True

################################################################################
def get_proxy():
    """ Adapted from https://stackoverflow.com/questions/59409418/how-to-rotate-selenium-webrowser-ip-address """
    options = Options()
    options.headless = True
    options.add_argument("window-size=700,600")
    driver = webdriver.Chrome(ChromeDriverManager().install(), options=options)
    clear_output()
    
    try:
        driver.get("https://sslproxies.org/")
        # on some machines the xpath is "//table[@class='table table-striped table-bordered dataTable']"???
        # table = driver.find_element_by_xpath("//table[@class='table table-striped table-bordered']")
        table = driver.find_elements_by_tag_name("table")[0]
        df = pd.read_html(table.get_attribute("outerHTML"))[0]
        df = df.iloc[np.where(~np.isnan(df["Port"]))[0],:] # ignore nans

        ips = df["IP Address"].values
        ports = df["Port"].astype("int").values

        driver.quit()
        proxies = list()
        for i in range(len(ips)):
            proxies.append("{}:{}".format(ips[i], ports[i]))
        i = random.randint(0, len(proxies)-1)
        return proxies[i]
    except Exception as e:
        driver.close()
        driver.quit()
        raise e



from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.common.exceptions import StaleElementReferenceException
from webdriver_manager.chrome import ChromeDriverManager
import pandas as pd
from IPython.display import clear_output

from sportstat.shared_functions import *

class Capology():

    ############################################################################
    def __init__(self):
    #---------------------------------------------------------------------------
        options = Options()
        options.add_argument('--headless')
        options.add_argument('window-size=700,600')
        # Use proxy
        # don't load images
        prefs = {'profile.managed_default_content_settings.images': 2}
        options.add_experimental_option('prefs', prefs)
        # create driver
        self.driver = webdriver.Chrome(ChromeDriverManager().install(), options=options)
        clear_output()

        self.leagues = {
            'Bundesliga': 'de/1-bundesliga',
            '2.Bundesliga': '/de/2-bundesliga',
            'EPL': 'uk/premier-league',
            'EFL Championship': '/uk/championship',
            'Serie A': 'it/serie-a',
            'Serie B': 'it/serie-b',
            'La Liga': 'es/la-liga',
            'La Liga 2': 'es/la-liga-2',
            'Ligue 1': 'fr/ligue-1',
            'Ligue 2': 'fr/ligue-2',
            'Eredivisie': '/ne/eredivisie',
            'Primeira Liga': '/pt/primeira-liga',
            'Scottish PL': '/uk/scottish-premiership',
            'Super Lig': '/tr/super-lig',
            'Belgian 1st Division': 'be/first-division-a'
        }


    ############################################################################
    def close(self):
    #---------------------------------------------------------------------------
        self.driver.close()
        self.driver.quit()

    
    ############################################################################
    def scrape_salaries(self, year, league, currency):
    #---------------------------------------------------------------------------
        error, valid = check_season(year, league, 'Capology')
        if not valid:
            print(error)
            return -1
        elif currency not in ['eur', 'gbp', 'usd']:
            print('Currency must be one of "eur", "gbp", or "usd".')
            return -1

        
        league_url = 'https://www.capology.com/{}/salaries/{}-{}'.format(self.leagues[league], year-1, year)

        self.driver.get(league_url)

        # show all players on one page
        done = False
        while not done:
            try:
                all_btn = WebDriverWait(
                    self.driver, 
                    10,
                ).until(EC.element_to_be_clickable(
                    (By.LINK_TEXT, 'All')
                ))
                all_btn.click()
                done = True
            except StaleElementReferenceException:
                pass

        # select the currency
        currency_btn = WebDriverWait(
            self.driver, 
            10,
        ).until(EC.element_to_be_clickable(
            (By.ID, 'btn_{}'.format(currency))
        ))
        self.driver.execute_script('arguments[0].click()', currency_btn)

        # table to pandas df
        tbody_html = self.driver.find_element(By.ID, 'table')\
                .find_element(By.TAG_NAME, 'tbody')\
                .get_attribute('outerHTML')
        table_html = '<table>' + tbody_html + '</table>'
        df = pd.read_html(table_html)[0]
        if df.shape[1] == 13:
            df = df.drop(columns=[1])
            df.columns = [
                'Player', 'Weekly Gross', 'Annual Gross', 'Expiration', 'Length', 
                'Total Gross', 'Status', 'Pos. group', 'Pos.', 'Age', 'Country', 
                'Club'
            ]
        else:
            df.columns = [
                'Player', 'Weekly Gross', 'Annual Gross', 'Adj. Gross', 'Pos. group', 
                'Age', 'Country', 'Club'
            ]

        return df

    
    ############################################################################
    def scrape_payrolls(self, year, league, currency):
    #---------------------------------------------------------------------------
        error, valid = check_season(year, league, 'Capology')
        if not valid:
            print(error)
            return -1
        elif currency not in ['eur', 'gbp', 'usd']:
            print('Currency must be one of "eur", "gbp", or "usd".')
            return -1

        
        league_url = 'https://www.capology.com/{}/payrolls/{}-{}'.format(self.leagues[league], year-1, year)

        self.driver.get(league_url)

        # select the currency
        currency_btn = WebDriverWait(
            self.driver, 
            10,
        ).until(EC.element_to_be_clickable(
            (By.ID, 'btn_{}'.format(currency))
        ))
        self.driver.execute_script('arguments[0].click()', currency_btn)

        # table to pandas df
        table = WebDriverWait(
            self.driver, 
            10,
        ).until(EC.element_to_be_clickable(
            (By.ID, 'table')
        ))
        done = False
        while not done:
            df = pd.read_html(table.get_attribute('outerHTML'))[0]
            done = True if df.shape[0]>0 else False
        # df.columns = [
        #     'Club', 'Weekly Gross (000s)', 'Annual Gross (000s)', 'Inflcation-Adj. Gross (000s)',
        #     'Keeper (000s)', 'Defense (000s)', 'Midfield (000s)', 'Forward (000s)'
        # ]

        return df



import setuptools

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()
    
setuptools.setup(
    name = 'sportstat',
    version = '2.0.0',
    author = 'Bernard Ojengwa',
    author_email = 'bernardojengwa@gmail.com',
    description = 'Package for scraping soccer data from a variety of sources.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/ojengwa/ScraperFC',
    package_dir = {'':'code'},
    packages = setuptools.find_packages(where='code'),
    keywords = ["soccer","football","Premier League","Serie A",
                "La Liga","Bundesliga","Ligue 1",'web scraping',
                'data'],
    install_requires = ['selenium', 
                        'webdriver-manager',
                        'pandas',
                        'numpy',
                        'datetime',
                        'ipython',
                        "requests",
                        "bs4",
                        "lxml",
                        "tqdm"],
    classifiers = ['Programming Language :: Python :: 3',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: OS Independent'],
    python_requires = '>=3.6'
)
  


from datetime import datetime
from io import StringIO
from IPython.display import clear_output
import pandas as pd
import requests
import time

class ClubElo:        
        
    def scrape_team_on_date(self, team, date):
        """
        Date must be passed in YYYY-MM-DD format as a string
        """
        date = datetime.strptime(date, '%Y-%m-%d')
        
        # use ClubElo API to get team data as Pandas DataFrame
        url = 'http://api.clubelo.com/{}'.format(team)
        done = False
        while not done:
            try:
                r = requests.get(url)
                done = True
            except requests.exceptions.ConnectionError:
                done = False
        df = pd.read_csv(StringIO(r.text), sep=',')
        
        # find row that given date falls in
        for i in df.index:
            from_date = datetime.strptime(df.loc[i,'From'], '%Y-%m-%d')
            to_date = datetime.strptime(df.loc[i,'To'], '%Y-%m-%d')
            if (date>from_date and date<to_date) or date==from_date or date==to_date:
                return df.loc[i,'Elo']
        
        return -1 # return -1 if ELO not found for given date
        



from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.common.exceptions import TimeoutException
from webdriver_manager.chrome import ChromeDriverManager
from sportstat.shared_functions import *
from tqdm import tqdm
import requests
from bs4 import BeautifulSoup
import pandas as pd


class Transfermarkt():
    
    ############################################################################
    def __init__(self):
        options = Options()
        prefs = {'profile.managed_default_content_settings.images': 2} # don't load images to make faster
        options.add_experimental_option('prefs', prefs)
        self.driver = webdriver.Chrome(ChromeDriverManager().install(), options=options) # create driver

        
    ############################################################################
    def close(self):
        self.driver.close()
        self.driver.quit()
        
        
    ############################################################################
    def get_club_links(self, league, year):
        print("Gathering club links.")
        error, valid = check_season(year, league, 'Transfermarkt')
        if not valid:
            print(error)
            return -1
        
        competition_links = {
            'EPL': 'https://www.transfermarkt.us/premier-league/startseite/wettbewerb/GB1',
            'EFL Championship': 'https://www.transfermarkt.us/championship/startseite/wettbewerb/GB2',
            'EFL1': 'https://www.transfermarkt.us/league-one/startseite/wettbewerb/GB3',
            'EFL2': 'https://www.transfermarkt.us/league-two/startseite/wettbewerb/GB4',
            'Bundesliga': 'https://www.transfermarkt.us/bundesliga/startseite/wettbewerb/L1',
            '2.Bundesliga': 'https://www.transfermarkt.us/2-bundesliga/startseite/wettbewerb/L2',
            'Serie A': 'https://www.transfermarkt.us/serie-a/startseite/wettbewerb/IT1',
            'Serie B': 'https://www.transfermarkt.us/serie-b/startseite/wettbewerb/IT2',
            'La Liga': 'https://www.transfermarkt.us/laliga/startseite/wettbewerb/ES1',
            'La Liga 2': 'https://www.transfermarkt.us/laliga2/startseite/wettbewerb/ES2',
            'Ligue 1': 'https://www.transfermarkt.us/ligue-1/startseite/wettbewerb/FR1',
            'Ligue 2': 'https://www.transfermarkt.us/ligue-2/startseite/wettbewerb/FR2',
            'Eredivisie': 'https://www.transfermarkt.us/eredivisie/startseite/wettbewerb/NL1',
            'Scottish PL': 'https://www.transfermarkt.us/scottish-premiership/startseite/wettbewerb/SC1',
            'Super Lig': 'https://www.transfermarkt.us/super-lig/startseite/wettbewerb/TR1',
            'Jupiler Pro League': 'https://www.transfermarkt.us/jupiler-pro-league/startseite/wettbewerb/BE1',
            'Liga Nos': 'https://www.transfermarkt.us/liga-nos/startseite/wettbewerb/PO1',
            'Russian Premier League': 'https://www.transfermarkt.us/premier-liga/startseite/wettbewerb/RU1',
            'Brasileirao': 'https://www.transfermarkt.us/campeonato-brasileiro-serie-a/startseite/wettbewerb/BRA1',
            'Argentina Liga Profesional': 'https://www.transfermarkt.us/superliga/startseite/wettbewerb/AR1N',
            'MLS': 'https://www.transfermarkt.us/major-league-soccer/startseite/wettbewerb/MLS1'
        }
        
        # go to the selected year
        url = '{}/plus/?saison_id={}'.format(competition_links[league], year-1)
        self.driver.get(url)
        
        # get the club links
        club_links = set()
        table = self.driver.find_elements(By.CLASS_NAME, 'items')[0]
        for el in table.find_elements(By.TAG_NAME, 'td'):
            if 'hauptlink no-border-links' in el.get_attribute('class'):
                href = el.find_element(By.TAG_NAME, 'a').get_attribute('href')
                club_links.add(href)
            
        return club_links
    
    
    ############################################################################
    def get_player_links(self, league, year):
        print("Gathering player links.")
        error, valid = check_season(year, league, 'Transfermarkt')
        if not valid:
            print(error)
            return -1
        
        player_links = set()
        club_links = self.get_club_links(league, year)
        for club_link in tqdm(club_links):
            self.driver.get(club_link)
            
            # get players from the table on the club page
            table = self.driver.find_elements(By.CLASS_NAME, 'items')[0]
            for el in table.find_elements(By.CLASS_NAME, 'hauptlink'):
                try:
                    subel = WebDriverWait(el, 10).until(EC.element_to_be_clickable((By.TAG_NAME, 'a')))
                except TimeoutException:
                    # Ben White f*cked everything up. For some reason, this works
                    continue
                href = subel.get_attribute('href')
                if 'profil' in href:
                    player_links.add(href)
                
        return player_links
    
    ############################################################################
    def get_players(self, league, year):
        error, valid = check_season(year, league, 'Transfermarkt')
        if not valid:
            print(error)
            return -1
        
        player_links = self.get_player_links(league, year)
        df = pd.DataFrame()
        for player_link in tqdm(player_links):
            player = TransfermarktPlayer(player_link)
            new_row = pd.Series(dtype=object)
            new_row["Name"]                 = player.name
            new_row["Value"]                = player.value
            new_row["Value last updated"]   = player.value_last_updated
            new_row["DOB"]                  = player.dob
            new_row["Age"]                  = player.age
            new_row["Height (m)"]           = player.height_meters
            new_row["Nationality"]          = player.nationality
            new_row["Citizenship"]          = player.citizenship
            new_row["Position"]             = player.position
            if player.other_positions is None:
                new_row["Other positions"]  = None
            else:
                new_row["Other positions"]  = pd.DataFrame(player.other_positions)
            new_row["Team"]                 = player.team
            new_row["Joined"]               = player.joined
            new_row["Contract expires"]     = player.contract_expires
            new_row["Market value history"] = player.market_value_history
            new_row["Transfer history"]     = player.transfer_history
            df = df.append(new_row, ignore_index=True)
        return df
    

################################################################################
class TransfermarktPlayer():
    
    def __init__(self, url):
        self.url = url
        headers = {
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '+\
                'AppleWebKit/537.36 (KHTML, like Gecko) '+\
                'Chrome/55.0.2883.87 Safari/537.36'
        }
        response = requests.get(url, headers=headers)
        soup = BeautifulSoup(response.content, 'html.parser')
        """ Name """
        data_header_el = soup.find("h1", {"class": "data-header__headline-wrapper"})
        self.name = data_header_el.getText().split('\n')[-1].strip()
        """ Value """
        try:
            self.value = soup.find("a", {"class": "data-header__market-value-wrapper"}).text.split(" ")[0]
            self.value_last_updated = soup.find("a", {"class": "data-header__market-value-wrapper"}).text.split("Last update: ")[-1]
        except AttributeError:
            self.value = None
            self.value_last_updated = None
        """ DOB and age """
        dob_el = soup.find("span", {"itemprop": "birthDate"})
        self.dob = ' '.join(dob_el.getText().strip().split(' ')[:3])
        self.age = int(dob_el.getText().strip().split(' ')[-1].replace('(','').replace(')',''))
        """ Height """
        height_el = soup.find("span", {"itemprop": "height"})
        try:
            self.height_meters = float(height_el.getText().replace("m", "").replace(",", "."))
        except AttributeError:
            self.height_meters = None
        """ Citizenship """
        nationality_el = soup.find("span", {"itemprop": "nationality"})
        self.nationality = nationality_el.getText().replace("\n","").strip()
        citizenship_els = soup.find_all("span", {"class": "info-table__content info-table__content--bold"})
        flag_els = [flag_el for el in citizenship_els\
            for flag_el in el.find_all("img", {"class": "flaggenrahmen"})]
        self.citizenship = list(set([el["title"] for el in flag_els]))
        """ Position """
        self.position = soup.find("dd", {"class": "detail-position__position"}).getText()
        try:
            self.other_positions = [el.getText() for el in soup.find("div", {"class": "detail-position__position"}).find_all("dd")]
        except AttributeError:
            self.other_positions = None
        """ Team & Contract """
        try:
            self.team = soup.find("span", {"class": "data-header__club"}).find("a")["title"]
        except TypeError:
            self.team = "Retired"
        if self.team in ["Without Club", "Retired"]:
            self.joined = None
            self.contract_expires = None
        else:
            self.joined = [el.text.split(": ")[-1] \
                for el in soup.find_all("span", {"class": "data-header__label"}) \
                if "joined" in el.text.lower()][0]
            self.contract_expires = [el.text.split(": ")[-1] \
                for el in soup.find_all("span", {"class": "data-header__label"}) \
                if "expires" in el.text.lower()][0]
        """ Market value history """
        try:
            script = [s for s in soup.find_all("script", {"type": "text/javascript"}) \
                if "var chart = new Highcharts.Chart" in str(s)][0]
            values = [int(s.split(",")[0]) for s in str(script).split("y':")[2:-2]]
            dates = [s.split("datum_mw':")[-1].split(",'x")[0].replace("\\x20"," ").replace("'", "") \
                 for s in str(script).split("y':")[2:-2]]
            self.market_value_history = pd.DataFrame({"date": dates, "value": values})
        except IndexError:
            self.market_value_history = None
        """ Transfer History """
        rows = soup.find_all("div", {"class": "tm-player-transfer-history-grid"})
        self.transfer_history = pd.DataFrame(columns=["Season", "Date", "Left", "Joined", "MV", "Fee"])
        reached_prev_transfers = True # Assume we've reached past transfers
        for row in rows[1:-1]:
            fields = [s.strip() for s in row.getText().split("\n\n") if s!=""]
            if "Upcoming transfer" in fields:
                # The next row will be a future transfer
                reached_prev_transfers = False
            if reached_prev_transfers:
                new_row = pd.Series({
                    "Season": fields[0],
                    "Date": fields[1],
                    "Left": fields[2],
                    "Joined": fields[3],
                    "MV": fields[4],
                    "Fee": fields[5]
                })
                self.transfer_history = self.transfer_history.append(new_row, ignore_index=True)
            if "Transfer history" in fields:
                # Now we've reached past transfers
                reached_prev_transfers = True
        